/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2009 EADS DEFENCE AND SECURITY SYSTEMS
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 */

package org.ow2.weblab.services.crawler.folder.impl;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebResult;
import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;

import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.content.binary.BinaryFolderContentManager;
import org.ow2.weblab.crawler.FolderCrawler;
import org.ow2.weblab.crawler.filter.ExtensionFilter;
import org.ow2.weblab.services.foldercrawling.FolderCrawling;
import org.ow2.weblab.services.foldercrawling.StartCrawlException;
import org.weblab_project.core.exception.WebLabCheckedException;
import org.weblab_project.core.exception.WebLabUncheckedException;
import org.weblab_project.core.factory.ResourceFactory;
import org.weblab_project.core.helper.PoKHelper;
import org.weblab_project.core.helper.RDFHelperFactory;
import org.weblab_project.core.model.ResourceCollection;
import org.weblab_project.core.ontologies.WebLab;
import org.weblab_project.core.properties.PropertiesLoader;
import org.weblab_project.services.configurable.Configurable;
import org.weblab_project.services.configurable.ConfigureException;
import org.weblab_project.services.configurable.ResetConfigurationException;
import org.weblab_project.services.configurable.types.ConfigureArgs;
import org.weblab_project.services.configurable.types.ConfigureReturn;
import org.weblab_project.services.configurable.types.ResetConfigurationArgs;
import org.weblab_project.services.configurable.types.ResetConfigurationReturn;
import org.weblab_project.services.exception.WebLabException;
import org.weblab_project.services.foldercrawling.types.StartCrawlArgs;
import org.weblab_project.services.foldercrawling.types.StartCrawlReturn;
import org.weblab_project.services.sourcereader.GetResourceException;
import org.weblab_project.services.sourcereader.SourceReader;
import org.weblab_project.services.sourcereader.types.GetResourceArgs;
import org.weblab_project.services.sourcereader.types.GetResourceReturn;

/**
 * @author jdoucy, ymombrun
 * @date 5 juin 08
 */
@WebService()
@SOAPBinding(parameterStyle = SOAPBinding.ParameterStyle.BARE)
public class FolderCrawlerService implements FolderCrawling, Configurable,
		SourceReader {

	/**
	 * Property denoting the folder to crawl
	 */
	private static final String FOLDER_PROPERTY = WebLab.PROCESSING_PROPERTY_NAMESPACE
			+ "crawler/folder";

	/**
	 * BinaryFolderContentManager
	 */
	protected BinaryFolderContentManager contentManager = null;

	/**
	 * The default folder to crawl
	 */
	public static final String DEFAULT_FOLDER = "toBeCrawled";

	/**
	 * The file defining the folder to crawl, in a one-a-row style
	 */
	public static final String FOLDERS_FILE = "FolderCrawlerService.config";

	private static final String FOLDERS = "folders";

	private static final String EXTENSIONS = "extensions";

	private static final String REJECT = "reject";

	private static final String RECURSIVE = "recursive";

	protected Map<String, String> props = null;

	protected Map<String, List<String>> usContextAndFolderToCrawl = new HashMap<String, List<String>>();

	protected Map<String, List<FolderCrawler>> usContextAndFolderCrawler = new HashMap<String, List<FolderCrawler>>();

	protected List<FolderCrawler> defaultFC = new ArrayList<FolderCrawler>();

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.weblab_project.services.foldercrawling.FolderCrawling#startCrawl(
	 * org.weblab_project.services.foldercrawling.types.StartCrawlArgs)
	 */
	@WebMethod(action = "startCrawl")
	@WebResult(name = "startCrawlReturn", targetNamespace = "http://weblab-project.org/services/foldercrawling/types", partName = "return")
	public StartCrawlReturn startCrawl(
			@WebParam(name = "startCrawlArgs", targetNamespace = "http://weblab-project.org/services/foldercrawling/types", partName = "args") final StartCrawlArgs args)
			throws StartCrawlException {

		LogFactory.getLog(this.getClass()).info("StartCrawl method called.");
		if (this.contentManager == null) {
			try {
				this.contentManager = BinaryFolderContentManager.getInstance();
			} catch (final WebLabUncheckedException wlue) {
				WebLabException exp = new WebLabException();
				exp.setErrorId("E0");
				exp.setErrorMessage("Unexpected error");
				throw new StartCrawlException("Unable to "
						+ "get the content manager.", exp, wlue);
			}
		}

		// Get the usageContext URI to be used as key
		String uc = null;
		if (args != null && args.getUsageContext() != null) {
			uc = args.getUsageContext().getUri();
		}

		if (uc == null) {
			this.defaultFC.clear();
			// if (this.defaultFC.isEmpty()) {
			// If there no usageContext in param --> load default
			LogFactory.getLog(this.getClass()).info(
					"UsageContext was null. "
							+ "Folder will be loaded from config file.");
			final List<String> filesToCrawl = this.getFoldersFromConfig();
			final boolean recursive = this.getRecursiveFromConfig();
			final FileFilter ff = this.getFileFilterFromConfig();
			for (final String toCrawl : filesToCrawl) {
				LogFactory.getLog(this.getClass()).info(
						"Create folder crawler on: " + toCrawl);
				FolderCrawler fc;
				try {
					fc = new FolderCrawler(this.contentManager, new File(
							toCrawl), ff, recursive);
				} catch (final WebLabCheckedException wlce) {
					LogFactory.getLog(this.getClass()).warn(
							"Unable to crawl folder: " + toCrawl, wlce);
					continue;
				}
				this.defaultFC.add(fc);
			}
			for (final FolderCrawler fc : this.defaultFC) {
				fc.startCrawl();
			}
			// } else {
			// LogFactory.getLog(this.getClass()).info("Default Folder to crawl was not empty. Nothing done.");
			// }
		} else if (!this.usContextAndFolderToCrawl.containsKey(uc)) {
			// If usageContext not configured --> load default
			LogFactory.getLog(this.getClass()).info(
					"No configure done with this usageContext( " + uc
							+ "). Folder will be loaded from config file.");
			final List<String> filesToCrawl = this.getFoldersFromConfig();
			final boolean recursive = this.getRecursiveFromConfig();
			final FileFilter ff = this.getFileFilterFromConfig();
			List<FolderCrawler> fcList = this.usContextAndFolderCrawler.get(uc);
			if (fcList == null) {
				fcList = new ArrayList<FolderCrawler>();
			}

			for (final String toCrawl : filesToCrawl) {
				FolderCrawler fc;
				try {
					fc = new FolderCrawler(this.contentManager, new File(
							toCrawl), ff, recursive);
				} catch (final WebLabCheckedException wlce) {
					LogFactory.getLog(this.getClass()).warn(
							"Unable to crawl folder: " + toCrawl, wlce);
					continue;
				}
				fc.startCrawl();
				fcList.add(fc);
			}
			this.usContextAndFolderCrawler.put(uc, fcList);

		} else {
			List<String> filesToCrawl = this.usContextAndFolderToCrawl.get(uc);
			if (filesToCrawl == null || filesToCrawl.isEmpty()) {
				LogFactory.getLog(this.getClass()).warn(
						"No folder to crawl associated with the usageContext ("
								+ uc + ")");
			} else {
				List<FolderCrawler> fcList = this.usContextAndFolderCrawler
						.get(uc);
				if (fcList == null) {
					fcList = new ArrayList<FolderCrawler>();
				}
				final boolean recursive = this.getRecursiveFromConfig();
				final FileFilter ff = this.getFileFilterFromConfig();
				for (final String toCrawl : filesToCrawl) {
					FolderCrawler fc;
					try {
						fc = new FolderCrawler(this.contentManager, new File(
								toCrawl), ff, recursive);
					} catch (final WebLabCheckedException wlce) {
						LogFactory.getLog(this.getClass()).warn(
								"Unable to crawl folder: " + toCrawl, wlce);
						continue;
					}
					fc.startCrawl();
					fcList.add(fc);
				}
				this.usContextAndFolderCrawler.put(uc, fcList);
			}
		}
		return new StartCrawlReturn();
	}

	private boolean getRecursiveFromConfig() {
		if (this.props == null) {
			this.loadProps();
		}
		final boolean rec;
		if (this.props.containsKey(FolderCrawlerService.RECURSIVE)) {
			rec = Boolean.parseBoolean(this.props
					.get(FolderCrawlerService.RECURSIVE));
		} else {
			LogFactory.getLog(this.getClass()).warn(
					"Unable to load recursive from config file. "
							+ "false will be used.");
			rec = false;
		}
		return rec;
	}

	private FileFilter getFileFilterFromConfig() {
		if (this.props == null) {
			this.loadProps();
		}
		final boolean rej;
		if (this.props.containsKey(FolderCrawlerService.REJECT)) {
			rej = Boolean.parseBoolean(this.props
					.get(FolderCrawlerService.REJECT));

		} else {
			LogFactory.getLog(this.getClass()).warn(
					"Unable to load reject from config file. "
							+ "false will be used.");
			rej = false;
		}
		List<String> ext = new ArrayList<String>();
		if (this.props.containsKey(FolderCrawlerService.EXTENSIONS)) {
			String[] splitted = this.props.get(FolderCrawlerService.EXTENSIONS)
					.split(";");
			for (String string : splitted) {
				string = string.trim();
				if (!string.equals("")) {
					ext.add(string);
				}
			}
			if (ext.isEmpty()) {
				LogFactory.getLog(this.getClass()).info(
						"Extensions is empty; from "
								+ FolderCrawlerService.FOLDERS_FILE);
			}
		} else {
			LogFactory.getLog(this.getClass()).warn(
					"Unable to get extensions from "
							+ FolderCrawlerService.FOLDERS_FILE);
		}
		return new ExtensionFilter(ext, rej);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.weblab_project.services.sourcereader.SourceReader#getResource(org
	 * .weblab_project.services.sourcereader.types.GetResourceArgs)
	 */
	@WebMethod(action = "getResource")
	@WebResult(name = "getResourceReturn", targetNamespace = "http://weblab-project.org/services/sourcereader/types", partName = "return")
	public GetResourceReturn getResource(
			@WebParam(name = "getResourceArgs", targetNamespace = "http://weblab-project.org/services/sourcereader/types", partName = "args") GetResourceArgs args)
			throws GetResourceException {
		int limit = -1;
		int offset = 0;

		if (args != null) {
			limit = args.getLimit();
			offset = args.getOffset();
		}
		LogFactory.getLog(this.getClass()).info(
				"GetCrawledDocuments method called. " + "offset: " + offset
						+ "; limit: " + limit);
		GetResourceReturn ret = new GetResourceReturn();
		if (this.contentManager == null) {
			try {
				this.contentManager = BinaryFolderContentManager.getInstance();
			} catch (final WebLabUncheckedException wlue) {
				WebLabException exp = new WebLabException();
				exp.setErrorId("E0");
				exp.setErrorMessage("Unexpected error");
				throw new GetResourceException(
						"Unable to get the content manager", exp, wlue);
			}
		}
		// Get the usageContext URI to be used as key
		String uc = null;
		if (args != null && args.getUsageContext() != null) {
			uc = args.getUsageContext().getUri();
		}
		List<FolderCrawler> fcList;
		if (uc == null) {
			fcList = this.defaultFC;
		} else if (this.usContextAndFolderCrawler.containsKey(uc)) {
			fcList = this.usContextAndFolderCrawler.get(uc);
		} else {
			LogFactory.getLog(this.getClass()).warn(
					"StartCrawl must have been called with this usageContext ("
							+ uc + ") before.");
			WebLabException exp = new WebLabException();
			exp.setErrorId("E5");
			exp.setErrorMessage("Unsupported request");
			throw new GetResourceException(
					"StartCrawl must have been called before.", exp);
		}

		ResourceCollection resCol = getRCFromList(fcList, offset, limit);

		if (resCol.getResource().size() == 0) {
			LogFactory.getLog(this.getClass()).info(
					"No resources in the resource collection "
							+ "returned by the crawler...");
			LogFactory.getLog(this.getClass()).info(
					"FoldersCrawler were: " + fcList);
		}
		ret.setResources(resCol);
		return ret;
	}

	private static ResourceCollection getRCFromList(List<FolderCrawler> fcList,
			int offset, int limit) {
		if (fcList == null || fcList.isEmpty()) {
			LogFactory.getLog(FolderCrawlerService.class).warn(
					"No folder to crawl");
			return ResourceFactory.createResource("folderCrawlerService",
					"emptyCollection" + System.currentTimeMillis(),
					ResourceCollection.class);
		}

		final int usedOffset;
		if (offset < 0) {
			usedOffset = 0;
		} else {
			usedOffset = offset;
		}
		final int usedLimit;
		if (limit <= 0) {
			usedLimit = Integer.MAX_VALUE;
		} else {
			usedLimit = limit;
		}

		int cpt = 0;
		ResourceCollection resCol = ResourceFactory.createResource(
				"folderCrawlerService", "tempCollection-"
						+ System.currentTimeMillis(), ResourceCollection.class);
		for (final FolderCrawler fc : fcList) {
			if (cpt + fc.getNbFiles() < usedOffset) {
				// fc is before the offset
				cpt += fc.getNbFiles();
				continue;
			}
			if (cpt >= usedOffset + usedLimit) {
				break;
			}
			resCol.getResource().addAll(
					fc.getCrawledDocuments(usedOffset - cpt, usedLimit)
							.getResource());
			cpt += fc.getNbFiles();
		}

		return resCol;
	}

	protected void loadProps() {
		Set<String> toLoad = new HashSet<String>();
		toLoad.add(FolderCrawlerService.EXTENSIONS);
		toLoad.add(FolderCrawlerService.FOLDERS);
		toLoad.add(FolderCrawlerService.RECURSIVE);
		toLoad.add(FolderCrawlerService.REJECT);
		this.props = PropertiesLoader.loadProperties(
				FolderCrawlerService.FOLDERS_FILE, toLoad);
	}

	/**
	 * @return The folder names listed in the file
	 *         FolderCrawlerService.FOLDERS_FILE
	 */
	private List<String> getFoldersFromConfig() {
		if (this.props == null) {
			this.loadProps();
		}
		List<String> folders = new ArrayList<String>();
		if (this.props.containsKey(FOLDERS)) {
			String[] splitted = this.props.get(FOLDERS).split(";");
			for (String string : splitted) {
				string = string.trim();
				if (!string.equals("")) {
					folders.add(string);
				}
			}
			if (folders.isEmpty()) {
				LogFactory.getLog(this.getClass()).warn(
						"Unable to get folders from "
								+ FolderCrawlerService.FOLDERS_FILE + ". ( "
								+ this.props.get(FolderCrawlerService.FOLDERS)
								+ ") Default folder will be used.");
				folders.add(FolderCrawlerService.DEFAULT_FOLDER);
			}
		} else {
			LogFactory.getLog(this.getClass()).warn(
					"Unable to load folders from "
							+ FolderCrawlerService.FOLDERS_FILE
							+ ". Default folder will be used.");
			folders.add(FolderCrawlerService.DEFAULT_FOLDER);
		}
		return folders;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.weblab_project.services.configurable.Configurable#configure(org.
	 * weblab_project.services.configurable.types.ConfigureArgs)
	 */
	@WebMethod(action = "configure")
	@WebResult(name = "configureReturn", targetNamespace = "http://weblab-project.org/services/configurable/types", partName = "return")
	public ConfigureReturn configure(
			@WebParam(name = "configureArgs", targetNamespace = "http://weblab-project.org/services/configurable/types", partName = "args") final ConfigureArgs args)
			throws ConfigureException {
		LogFactory.getLog(this.getClass()).info("Configure method called.");
		if (args == null || args.getUsageContext() == null
				|| args.getConfiguration() == null
				|| args.getUsageContext().getUri() == null) {
			WebLabException wle = new WebLabException();
			wle.setErrorId("E1");
			wle.setErrorMessage("Invalid parameter");
			throw new ConfigureException(
					"ConfigureArgs was invalid (either it self, "
							+ "usageContext, it's uri "
							+ "or configuration was null", wle);
		}
		final PoKHelper helper = RDFHelperFactory.getPoKHelper(args
				.getConfiguration());
		List<String> list = new ArrayList<String>();
		list.addAll(helper.getLitsOnPredSubj(args.getUsageContext().getUri(),
				FolderCrawlerService.FOLDER_PROPERTY));
		list.addAll(helper.getRessOnPredSubj(args.getUsageContext().getUri(),
				FolderCrawlerService.FOLDER_PROPERTY));

		if (list.isEmpty()) {
			WebLabException exp = new WebLabException();
			exp.setErrorId("E1");
			exp.setErrorMessage("Unable to read configuration annotation.");
			throw new ConfigureException("No property '"
					+ FolderCrawlerService.FOLDER_PROPERTY
					+ "' in the configuration annotation.", exp);
		}

		this.usContextAndFolderToCrawl.put(args.getUsageContext().getUri(),
				list);

		return new ConfigureReturn();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.weblab_project.services.configurable.Configurable#resetConfiguration
	 * (org.weblab_project.services.configurable.types.ResetConfigurationArgs)
	 */
	@WebMethod(action = "resetConfiguration")
	@WebResult(name = "resetConfigurationReturn", targetNamespace = "http://weblab-project.org/services/configurable/types", partName = "return")
	public ResetConfigurationReturn resetConfiguration(
			@WebParam(name = "resetConfigurationArgs", targetNamespace = "http://weblab-project.org/services/configurable/types", partName = "args") final ResetConfigurationArgs args)
			throws ResetConfigurationException {
		LogFactory.getLog(this.getClass()).info(
				"ResetConfiguration method called.");
		this.defaultFC.clear();
		this.usContextAndFolderCrawler.clear();
		this.usContextAndFolderToCrawl.clear();
		return new ResetConfigurationReturn();
	}

}
